from __future__ import absolute_import

import sys

import abc
import numpy as np
import tensorflow as tf
import tensorflow.contrib.slim as slim

from tf_pose.common import to_str
from tf_pose import common

DEFAULT_PADDING = "SAME"


_init_xavier = tf.contrib.layers.xavier_initializer()
_init_norm = tf.truncated_normal_initializer(stddev=0.01)
_init_zero = slim.init_ops.zeros_initializer()
_l2_regularizer_00004 = tf.contrib.layers.l2_regularizer(0.00004)
_l2_regularizer_convb = tf.contrib.layers.l2_regularizer(common.regularizer_conv)


def layer(op):
    """
    Decorator for composable network layers.
    """

    def layer_decorated(self, *args, **kwargs):
        # Automatically set a name if not provided.
        name = kwargs.setdefault("name", self.get_unique_name(op.__name__))
        # Figure out the layer inputs.
        if len(self.terminals) == 0:
            raise RuntimeError("No input variables found for layer %s." % name)
        elif len(self.terminals) == 1:
            layer_input = self.terminals[0]
        else:
            layer_input = list(self.terminals)
        # Perform the operation and get the output.
        layer_output = op(self, layer_input, *args, **kwargs)
        # Add to layer LUT.
        self.layers[name] = layer_output
        # This output is now the input for the next layer.
        self.feed(layer_output)
        # Return self for chained calls.
        return self

    return layer_decorated


class BaseNetwork(object):
    def __init__(self, inputs, trainable=True):
        # The input nodes for this network
        self.inputs = inputs
        # The current list of terminal nodes
        self.terminals = []
        # Mapping from layer names to layers
        self.layers = dict(inputs)
        # If true, the resulting variables are set as trainable
        self.trainable = trainable
        # Switch variable for dropout
        self.use_dropout = tf.placeholder_with_default(
            tf.constant(1.0), shape=[], name="use_dropout"
        )
        self.setup()

    @abc.abstractmethod
    def setup(self):
        """Construct the network. """
        raise NotImplementedError("Must be implemented by the subclass.")

    def load(self, data_path, session, ignore_missing=False):
        """
        Load network weights.
        data_path: The path to the numpy-serialized network weights
        session: The current TensorFlow session
        ignore_missing: If true, serialized weights for missing layers are ignored.
        """
        data_dict = np.load(data_path, encoding="bytes").item()
        for op_name, param_dict in data_dict.items():
            if isinstance(data_dict[op_name], np.ndarray):
                if "RMSProp" in op_name:
                    continue
                with tf.variable_scope("", reuse=True):
                    var = tf.get_variable(op_name.replace(":0", ""))
                    try:
                        session.run(var.assign(data_dict[op_name]))
                    except Exception as e:
                        print(op_name)
                        print(e)
                        sys.exit(-1)
            else:
                op_name = to_str(op_name)
                # if op_name > 'conv4':
                #     print(op_name, 'skipped')
                #     continue
                # print(op_name, 'restored')
                with tf.variable_scope(op_name, reuse=True):
                    for param_name, data in param_dict.items():
                        try:
                            var = tf.get_variable(to_str(param_name))
                            session.run(var.assign(data))
                        except ValueError as e:
                            print(e)
                            if not ignore_missing:
                                raise

    def feed(self, *args):
        """Set the input(s) for the next operation by replacing the terminal nodes.
        The arguments can be either layer names or the actual layers.
        """
        assert len(args) != 0
        self.terminals = []
        for fed_layer in args:
            try:
                is_str = isinstance(fed_layer, basestring)
            except NameError:
                is_str = isinstance(fed_layer, str)
            if is_str:
                try:
                    fed_layer = self.layers[fed_layer]
                except KeyError:
                    raise KeyError("Unknown layer name fed: %s" % fed_layer)
            self.terminals.append(fed_layer)
        return self

    def get_output(self, name=None):
        """Returns the current network output."""
        if not name:
            return self.terminals[-1]
        else:
            return self.layers[name]

    def get_tensor(self, name):
        return self.get_output(name)

    def get_unique_name(self, prefix):
        """Returns an index-suffixed unique name for the given prefix.
        This is used for auto-generating layer names based on the type-prefix.
        """
        ident = sum(t.startswith(prefix) for t, _ in self.layers.items()) + 1
        return "%s_%d" % (prefix, ident)

    def make_var(self, name, shape, trainable=True):
        """Creates a new TensorFlow variable."""
        return tf.get_variable(
            name,
            shape,
            trainable=self.trainable & trainable,
            initializer=tf.contrib.layers.xavier_initializer(),
        )

    def validate_padding(self, padding):
        """Verifies that the padding is one of the supported ones."""
        assert padding in ("SAME", "VALID")

    @layer
    def normalize_vgg(self, input, name):
        # normalize input -0.5 ~ 0.5
        input = tf.multiply(input, 1.0 / 256.0, name=name + "_divide")
        input = tf.add(input, -0.5, name=name + "_subtract")
        return input

    @layer
    def normalize_mobilenet(self, input, name):
        input = tf.divide(input, 128.0, name=name + "_divide")
        input = tf.subtract(input, 1.0, name=name + "_subtract")
        return input

    @layer
    def normalize_nasnet(self, input, name):
        input = tf.divide(input, 255.0, name=name + "_divide")
        input = tf.subtract(input, 0.5, name=name + "_subtract")
        input = tf.multiply(input, 2.0, name=name + "_multiply")
        return input

    @layer
    def upsample(self, input, factor, name):
        if isinstance(factor, str):
            sh = tf.shape(self.get_tensor(factor))[1:3]
        else:
            sh = tf.shape(input)[1:3] * factor
        return tf.image.resize_bilinear(input, sh, align_corners=False, name=name)

    @layer
    def separable_conv(
        self, input, k_h, k_w, c_o, stride, name, relu=True, set_bias=True
    ):
        with slim.arg_scope(
            [slim.batch_norm],
            decay=0.999,
            fused=common.batchnorm_fused,
            is_training=self.trainable,
        ):
            output = slim.separable_convolution2d(
                input,
                num_outputs=None,
                stride=stride,
                trainable=self.trainable,
                depth_multiplier=1.0,
                kernel_size=[k_h, k_w],
                # activation_fn=common.activation_fn if relu else None,
                activation_fn=None,
                # normalizer_fn=slim.batch_norm,
                weights_initializer=_init_xavier,
                # weights_initializer=_init_norm,
                weights_regularizer=_l2_regularizer_00004,
                biases_initializer=None,
                padding=DEFAULT_PADDING,
                scope=name + "_depthwise",
            )

            output = slim.convolution2d(
                output,
                c_o,
                stride=1,
                kernel_size=[1, 1],
                activation_fn=common.activation_fn if relu else None,
                weights_initializer=_init_xavier,
                # weights_initializer=_init_norm,
                biases_initializer=_init_zero if set_bias else None,
                normalizer_fn=slim.batch_norm,
                trainable=self.trainable,
                weights_regularizer=None,
                scope=name + "_pointwise",
            )

        return output

    @layer
    def convb(
        self,
        input,
        k_h,
        k_w,
        c_o,
        stride,
        name,
        relu=True,
        set_bias=True,
        set_tanh=False,
    ):
        with slim.arg_scope(
            [slim.batch_norm],
            decay=0.999,
            fused=common.batchnorm_fused,
            is_training=self.trainable,
        ):
            output = slim.convolution2d(
                input,
                c_o,
                kernel_size=[k_h, k_w],
                stride=stride,
                normalizer_fn=slim.batch_norm,
                weights_regularizer=_l2_regularizer_convb,
                weights_initializer=_init_xavier,
                # weights_initializer=tf.truncated_normal_initializer(stddev=0.01),
                biases_initializer=_init_zero if set_bias else None,
                trainable=self.trainable,
                activation_fn=common.activation_fn if relu else None,
                scope=name,
            )
            if set_tanh:
                output = tf.nn.sigmoid(output, name=name + "_extra_acv")
        return output

    @layer
    def conv(
        self,
        input,
        k_h,
        k_w,
        c_o,
        s_h,
        s_w,
        name,
        relu=True,
        padding=DEFAULT_PADDING,
        group=1,
        trainable=True,
        biased=True,
    ):
        # Verify that the padding is acceptable
        self.validate_padding(padding)
        # Get the number of channels in the input
        c_i = int(input.get_shape()[-1])
        # Verify that the grouping parameter is valid
        assert c_i % group == 0
        assert c_o % group == 0
        # Convolution for a given input and kernel
        convolve = lambda i, k: tf.nn.conv2d(i, k, [1, s_h, s_w, 1], padding=padding)
        with tf.variable_scope(name) as scope:
            kernel = self.make_var(
                "weights",
                shape=[k_h, k_w, c_i / group, c_o],
                trainable=self.trainable & trainable,
            )
            if group == 1:
                # This is the common-case. Convolve the input without any further complications.
                output = convolve(input, kernel)
            else:
                # Split the input into groups and then convolve each of them independently
                input_groups = tf.split(3, group, input)
                kernel_groups = tf.split(3, group, kernel)
                output_groups = [
                    convolve(i, k) for i, k in zip(input_groups, kernel_groups)
                ]
                # Concatenate the groups
                output = tf.concat(3, output_groups)
            # Add the biases
            if biased:
                biases = self.make_var(
                    "biases", [c_o], trainable=self.trainable & trainable
                )
                output = tf.nn.bias_add(output, biases)

            if relu:
                # ReLU non-linearity
                output = tf.nn.relu(output, name=scope.name)
            return output

    @layer
    def relu(self, input, name):
        return tf.nn.relu(input, name=name)

    @layer
    def max_pool(self, input, k_h, k_w, s_h, s_w, name, padding=DEFAULT_PADDING):
        self.validate_padding(padding)
        return tf.nn.max_pool(
            input,
            ksize=[1, k_h, k_w, 1],
            strides=[1, s_h, s_w, 1],
            padding=padding,
            name=name,
        )

    @layer
    def avg_pool(self, input, k_h, k_w, s_h, s_w, name, padding=DEFAULT_PADDING):
        self.validate_padding(padding)
        return tf.nn.avg_pool(
            input,
            ksize=[1, k_h, k_w, 1],
            strides=[1, s_h, s_w, 1],
            padding=padding,
            name=name,
        )

    @layer
    def lrn(self, input, radius, alpha, beta, name, bias=1.0):
        return tf.nn.local_response_normalization(
            input, depth_radius=radius, alpha=alpha, beta=beta, bias=bias, name=name
        )

    @layer
    def concat(self, inputs, axis, name):
        return tf.concat(axis=axis, values=inputs, name=name)

    @layer
    def add(self, inputs, name):
        return tf.add_n(inputs, name=name)

    @layer
    def fc(self, input, num_out, name, relu=True):
        with tf.variable_scope(name) as scope:
            input_shape = input.get_shape()
            if input_shape.ndims == 4:
                # The input is spatial. Vectorize it first.
                dim = 1
                for d in input_shape[1:].as_list():
                    dim *= d
                feed_in = tf.reshape(input, [-1, dim])
            else:
                feed_in, dim = (input, input_shape[-1].value)
            weights = self.make_var("weights", shape=[dim, num_out])
            biases = self.make_var("biases", [num_out])
            op = tf.nn.relu_layer if relu else tf.nn.xw_plus_b
            fc = op(feed_in, weights, biases, name=scope.name)
            return fc

    @layer
    def softmax(self, input, name):
        input_shape = map(lambda v: v.value, input.get_shape())
        if len(input_shape) > 2:
            # For certain models (like NiN), the singleton spatial dimensions
            # need to be explicitly squeezed, since they're not broadcast-able
            # in TensorFlow's NHWC ordering (unlike Caffe's NCHW).
            if input_shape[1] == 1 and input_shape[2] == 1:
                input = tf.squeeze(input, squeeze_dims=[1, 2])
            else:
                raise ValueError("Rank 2 tensor input expected for softmax!")
        return tf.nn.softmax(input, name=name)

    @layer
    def batch_normalization(self, input, name, scale_offset=True, relu=False):
        # NOTE: Currently, only inference is supported
        with tf.variable_scope(name) as scope:
            shape = [input.get_shape()[-1]]
            if scale_offset:
                scale = self.make_var("scale", shape=shape)
                offset = self.make_var("offset", shape=shape)
            else:
                scale, offset = (None, None)
            output = tf.nn.batch_normalization(
                input,
                mean=self.make_var("mean", shape=shape),
                variance=self.make_var("variance", shape=shape),
                offset=offset,
                scale=scale,
                # TODO: This is the default Caffe batch norm eps
                # Get the actual eps from parameters
                variance_epsilon=1e-5,
                name=name,
            )
            if relu:
                output = tf.nn.relu(output)
            return output

    @layer
    def dropout(self, input, keep_prob, name):
        keep = 1 - self.use_dropout + (self.use_dropout * keep_prob)
        return tf.nn.dropout(input, keep, name=name)

    @layer
    def se_block(self, input_feature, name, ratio=8):
        """Contains the implementation of Squeeze-and-Excitation block.
        As described in https://arxiv.org/abs/1709.01507.
        ref : https://github.com/kobiso/SENet-tensorflow-slim/blob/master/nets/attention_module.py
        """

        kernel_initializer = tf.contrib.layers.variance_scaling_initializer()
        bias_initializer = tf.constant_initializer(value=0.0)

        with tf.variable_scope(name):
            channel = input_feature.get_shape()[-1]
            # Global average pooling
            squeeze = tf.reduce_mean(input_feature, axis=[1, 2], keepdims=True)
            excitation = tf.layers.dense(
                inputs=squeeze,
                units=channel // ratio,
                activation=tf.nn.relu,
                kernel_initializer=kernel_initializer,
                bias_initializer=bias_initializer,
                name="bottleneck_fc",
            )
            excitation = tf.layers.dense(
                inputs=excitation,
                units=channel,
                activation=tf.nn.sigmoid,
                kernel_initializer=kernel_initializer,
                bias_initializer=bias_initializer,
                name="recover_fc",
            )
            scale = input_feature * excitation
        return scale



import math
import random

import cv2
import numpy as np
from tensorpack.dataflow.imgaug.geometry import RotationAndCropValid

from tf_pose.common import CocoPart

_network_w = 368
_network_h = 368
_scale = 2


def set_network_input_wh(w, h):
    global _network_w, _network_h
    _network_w, _network_h = w, h


def set_network_scale(scale):
    global _scale
    _scale = scale


def pose_random_scale(meta):
    scalew = random.uniform(0.8, 1.2)
    scaleh = random.uniform(0.8, 1.2)
    neww = int(meta.width * scalew)
    newh = int(meta.height * scaleh)
    dst = cv2.resize(meta.img, (neww, newh), interpolation=cv2.INTER_AREA)

    # adjust meta data
    adjust_joint_list = []
    for joint in meta.joint_list:
        adjust_joint = []
        for point in joint:
            if point[0] < -100 or point[1] < -100:
                adjust_joint.append((-1000, -1000))
                continue
            # if point[0] <= 0 or point[1] <= 0 or int(point[0] * scalew + 0.5) > neww or int(
            #                         point[1] * scaleh + 0.5) > newh:
            #     adjust_joint.append((-1, -1))
            #     continue
            adjust_joint.append(
                (int(point[0] * scalew + 0.5), int(point[1] * scaleh + 0.5))
            )
        adjust_joint_list.append(adjust_joint)

    meta.joint_list = adjust_joint_list
    meta.width, meta.height = neww, newh
    meta.img = dst
    return meta


def pose_resize_shortestedge_fixed(meta):
    ratio_w = _network_w / meta.width
    ratio_h = _network_h / meta.height
    ratio = max(ratio_w, ratio_h)
    return pose_resize_shortestedge(
        meta, int(min(meta.width * ratio + 0.5, meta.height * ratio + 0.5))
    )


def pose_resize_shortestedge_random(meta):
    ratio_w = _network_w / meta.width
    ratio_h = _network_h / meta.height
    ratio = min(ratio_w, ratio_h)
    target_size = int(min(meta.width * ratio + 0.5, meta.height * ratio + 0.5))
    target_size = int(target_size * random.uniform(0.95, 1.6))
    # target_size = int(min(_network_w, _network_h) * random.uniform(0.7, 1.5))
    return pose_resize_shortestedge(meta, target_size)


def pose_resize_shortestedge(meta, target_size):
    global _network_w, _network_h
    img = meta.img

    # adjust image
    scale = target_size / min(meta.height, meta.width)
    if meta.height < meta.width:
        newh, neww = target_size, int(scale * meta.width + 0.5)
    else:
        newh, neww = int(scale * meta.height + 0.5), target_size

    dst = cv2.resize(img, (neww, newh), interpolation=cv2.INTER_AREA)

    pw = ph = 0
    if neww < _network_w or newh < _network_h:
        pw = max(0, (_network_w - neww) // 2)
        ph = max(0, (_network_h - newh) // 2)
        mw = (_network_w - neww) % 2
        mh = (_network_h - newh) % 2
        color = random.randint(0, 255)
        dst = cv2.copyMakeBorder(
            dst, ph, ph + mh, pw, pw + mw, cv2.BORDER_CONSTANT, value=(color, 0, 0)
        )

    # adjust meta data
    adjust_joint_list = []
    for joint in meta.joint_list:
        adjust_joint = []
        for point in joint:
            if point[0] < -100 or point[1] < -100:
                adjust_joint.append((-1000, -1000))
                continue
            # if point[0] <= 0 or point[1] <= 0 or int(point[0]*scale+0.5) > neww or int(point[1]*scale+0.5) > newh:
            #     adjust_joint.append((-1, -1))
            #     continue
            adjust_joint.append(
                (int(point[0] * scale + 0.5) + pw, int(point[1] * scale + 0.5) + ph)
            )
        adjust_joint_list.append(adjust_joint)

    meta.joint_list = adjust_joint_list
    meta.width, meta.height = neww + pw * 2, newh + ph * 2
    meta.img = dst
    return meta


def pose_crop_center(meta):
    global _network_w, _network_h
    target_size = (_network_w, _network_h)
    x = (meta.width - target_size[0]) // 2 if meta.width > target_size[0] else 0
    y = (meta.height - target_size[1]) // 2 if meta.height > target_size[1] else 0

    return pose_crop(meta, x, y, target_size[0], target_size[1])


def pose_crop_random(meta):
    global _network_w, _network_h
    target_size = (_network_w, _network_h)

    for _ in range(50):
        x = (
            random.randrange(0, meta.width - target_size[0])
            if meta.width > target_size[0]
            else 0
        )
        y = (
            random.randrange(0, meta.height - target_size[1])
            if meta.height > target_size[1]
            else 0
        )

        # check whether any face is inside the box to generate a reasonably-balanced datasets
        for joint in meta.joint_list:
            if (
                x <= joint[CocoPart.Nose.value][0] < x + target_size[0]
                and y <= joint[CocoPart.Nose.value][1] < y + target_size[1]
            ):
                break

    return pose_crop(meta, x, y, target_size[0], target_size[1])


def pose_crop(meta, x, y, w, h):
    # adjust image
    target_size = (w, h)

    img = meta.img
    resized = img[y : y + target_size[1], x : x + target_size[0], :]

    # adjust meta data
    adjust_joint_list = []
    for joint in meta.joint_list:
        adjust_joint = []
        for point in joint:
            if point[0] < -100 or point[1] < -100:
                adjust_joint.append((-1000, -1000))
                continue
            # if point[0] <= 0 or point[1] <= 0:
            #     adjust_joint.append((-1000, -1000))
            #     continue
            new_x, new_y = point[0] - x, point[1] - y
            # if new_x <= 0 or new_y <= 0 or new_x > target_size[0] or new_y > target_size[1]:
            #     adjust_joint.append((-1, -1))
            #     continue
            adjust_joint.append((new_x, new_y))
        adjust_joint_list.append(adjust_joint)

    meta.joint_list = adjust_joint_list
    meta.width, meta.height = target_size
    meta.img = resized
    return meta


def pose_flip(meta):
    r = random.uniform(0, 1.0)
    if r > 0.5:
        return meta

    img = meta.img
    img = cv2.flip(img, 1)

    # flip meta
    flip_list = [
        CocoPart.Nose,
        CocoPart.Neck,
        CocoPart.LShoulder,
        CocoPart.LElbow,
        CocoPart.LWrist,
        CocoPart.RShoulder,
        CocoPart.RElbow,
        CocoPart.RWrist,
        CocoPart.LHip,
        CocoPart.LKnee,
        CocoPart.LAnkle,
        CocoPart.RHip,
        CocoPart.RKnee,
        CocoPart.RAnkle,
        CocoPart.LEye,
        CocoPart.REye,
        CocoPart.LEar,
        CocoPart.REar,
        CocoPart.Background,
    ]
    adjust_joint_list = []
    for joint in meta.joint_list:
        adjust_joint = []
        for cocopart in flip_list:
            point = joint[cocopart.value]
            if point[0] < -100 or point[1] < -100:
                adjust_joint.append((-1000, -1000))
                continue
            # if point[0] <= 0 or point[1] <= 0:
            #     adjust_joint.append((-1, -1))
            #     continue
            adjust_joint.append((meta.width - point[0], point[1]))
        adjust_joint_list.append(adjust_joint)

    meta.joint_list = adjust_joint_list

    meta.img = img
    return meta


def pose_rotation(meta):
    deg = random.uniform(-15.0, 15.0)
    img = meta.img

    center = (img.shape[1] * 0.5, img.shape[0] * 0.5)  # x, y
    rot_m = cv2.getRotationMatrix2D((int(center[0]), int(center[1])), deg, 1)
    ret = cv2.warpAffine(
        img,
        rot_m,
        img.shape[1::-1],
        flags=cv2.INTER_AREA,
        borderMode=cv2.BORDER_CONSTANT,
    )
    if img.ndim == 3 and ret.ndim == 2:
        ret = ret[:, :, np.newaxis]
    neww, newh = RotationAndCropValid.largest_rotated_rect(
        ret.shape[1], ret.shape[0], deg
    )
    neww = min(neww, ret.shape[1])
    newh = min(newh, ret.shape[0])
    newx = int(center[0] - neww * 0.5)
    newy = int(center[1] - newh * 0.5)
    # print(ret.shape, deg, newx, newy, neww, newh)
    img = ret[newy : newy + newh, newx : newx + neww]

    # adjust meta data
    adjust_joint_list = []
    for joint in meta.joint_list:
        adjust_joint = []
        for point in joint:
            if point[0] < -100 or point[1] < -100:
                adjust_joint.append((-1000, -1000))
                continue
            # if point[0] <= 0 or point[1] <= 0:
            #     adjust_joint.append((-1, -1))
            #     continue
            x, y = _rotate_coord((meta.width, meta.height), (newx, newy), point, deg)
            adjust_joint.append((x, y))
        adjust_joint_list.append(adjust_joint)

    meta.joint_list = adjust_joint_list
    meta.width, meta.height = neww, newh
    meta.img = img

    return meta


def _rotate_coord(shape, newxy, point, angle):
    angle = -1 * angle / 180.0 * math.pi

    ox, oy = shape
    px, py = point

    ox /= 2
    oy /= 2

    qx = math.cos(angle) * (px - ox) - math.sin(angle) * (py - oy)
    qy = math.sin(angle) * (px - ox) + math.cos(angle) * (py - oy)

    new_x, new_y = newxy

    qx += ox - new_x
    qy += oy - new_y

    return int(qx + 0.5), int(qy + 0.5)


def pose_to_img(meta_l):
    global _network_w, _network_h, _scale
    return [
        meta_l[0].img.astype(np.float16),
        meta_l[0].get_heatmap(target_size=(_network_w // _scale, _network_h // _scale)),
        meta_l[0].get_vectormap(
            target_size=(_network_w // _scale, _network_h // _scale)
        ),
    ]



__version__ = "0.11.0"
# from __future__ import absolute_import
# from __future__ import division
# from __future__ import print_function
#
# from tf_pose.runner import infer, Estimator, get_estimator



# vim: sta:et:sw=2:ts=2:sts=2
# Written by Antonio Loquercio

import numpy as np
import scipy.stats as st
import pdb

import tensorflow as tf


def layer(op):
    def layer_decorated(self, *args, **kwargs):
        # Automatically set a name if not provided.
        name = kwargs.setdefault("name", self.get_unique_name(op.__name__))
        # Figure out the layer inputs.
        if len(self.terminals) == 0:
            raise RuntimeError("No input variables found for layer %s." % name)
        elif len(self.terminals) == 1:
            layer_input = self.terminals[0]
        else:
            layer_input = list(self.terminals)
        # Perform the operation and get the output.
        layer_output = op(self, layer_input, *args, **kwargs)
        # Add to layer LUT.
        self.layers[name] = layer_output
        # This output is now the input for the next layer.
        self.feed(layer_output)
        # Return self for chained calls.
        return self

    return layer_decorated


class Smoother(object):
    def __init__(self, inputs, filter_size, sigma, heat_map_size=0):
        self.inputs = inputs
        self.terminals = []
        self.layers = dict(inputs)
        self.filter_size = filter_size
        self.sigma = sigma
        self.heat_map_size = heat_map_size
        self.setup()

    def setup(self):
        self.feed("data").conv(name="smoothing")

    def get_unique_name(self, prefix):
        ident = sum(t.startswith(prefix) for t, _ in self.layers.items()) + 1
        return "%s_%d" % (prefix, ident)

    def feed(self, *args):
        assert len(args) != 0
        self.terminals = []
        for fed_layer in args:
            if isinstance(fed_layer, str):
                try:
                    fed_layer = self.layers[fed_layer]
                except KeyError:
                    raise KeyError("Unknown layer name fed: %s" % fed_layer)
            self.terminals.append(fed_layer)
        return self

    def gauss_kernel(self, kernlen=21, nsig=3, channels=1):
        interval = (2 * nsig + 1.0) / (kernlen)
        x = np.linspace(-nsig - interval / 2.0, nsig + interval / 2.0, kernlen + 1)
        kern1d = np.diff(st.norm.cdf(x))
        kernel_raw = np.sqrt(np.outer(kern1d, kern1d))
        kernel = kernel_raw / kernel_raw.sum()
        out_filter = np.array(kernel, dtype=np.float32)
        out_filter = out_filter.reshape((kernlen, kernlen, 1, 1))
        out_filter = np.repeat(out_filter, channels, axis=2)
        return out_filter

    def make_gauss_var(self, name, size, sigma, c_i):
        # with tf.device("/cpu:0"):
        kernel = self.gauss_kernel(size, sigma, c_i)
        var = tf.Variable(tf.convert_to_tensor(kernel), name=name)
        return var

    def get_output(self):
        """Returns the smoother output."""
        return self.terminals[-1]

    @layer
    def conv(self, input, name, padding="SAME"):
        # Get the number of channels in the input
        if self.heat_map_size != 0:
            c_i = self.heat_map_size
        else:
            c_i = input.get_shape().as_list()[3]
        # Convolution for a given input and kernel
        convolve = lambda i, k: tf.nn.depthwise_conv2d(
            i, k, [1, 1, 1, 1], padding=padding
        )
        with tf.variable_scope(name) as scope:
            kernel = self.make_gauss_var(
                "gauss_weight", self.filter_size, self.sigma, c_i
            )
            output = convolve(input, kernel)
        return output



import time
from collections import defaultdict


class StopWatchManager:
    def __init__(self):
        self.watches = defaultdict(StopWatch)

    def get(self, name):
        return self.watches[name]

    def start(self, name):
        self.get(name).start()

    def stop(self, name):
        self.get(name).stop()

    def reset(self, name):
        self.get(name).reset()

    def get_elapsed(self, name):
        return self.get(name).get_elapsed()

    def __repr__(self):
        return "\n".join(
            ["%s: %.8f" % (k, v.elapsed_accumulated) for k, v in self.watches.items()]
        )


class StopWatch:
    def __init__(self):
        self.elapsed_accumulated = 0.0
        self.started_at = time.time()

    def start(self):
        self.started_at = time.time()

    def stop(self):
        self.elapsed_accumulated += time.time() - self.started_at

    def reset(self):
        self.elapsed_accumulated = 0.0

    def get_elapsed(self):
        return self.elapsed_accumulated



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: datum.proto

import sys

_b = sys.version_info[0] < 3 and (lambda x: x) or (lambda x: x.encode("latin1"))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
from google.protobuf import descriptor_pb2

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


DESCRIPTOR = _descriptor.FileDescriptor(
    name="datum.proto",
    package="",
    serialized_pb=_b(
        '\n\x0b\x64\x61tum.proto"\x81\x01\n\x05\x44\x61tum\x12\x10\n\x08\x63hannels\x18\x01 \x01(\x05\x12\x0e\n\x06height\x18\x02 \x01(\x05\x12\r\n\x05width\x18\x03 \x01(\x05\x12\x0c\n\x04\x64\x61ta\x18\x04 \x01(\x0c\x12\r\n\x05label\x18\x05 \x01(\x05\x12\x12\n\nfloat_data\x18\x06 \x03(\x02\x12\x16\n\x07\x65ncoded\x18\x07 \x01(\x08:\x05\x66\x61lse'
    ),
)
_sym_db.RegisterFileDescriptor(DESCRIPTOR)


_DATUM = _descriptor.Descriptor(
    name="Datum",
    full_name="Datum",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name="channels",
            full_name="Datum.channels",
            index=0,
            number=1,
            type=5,
            cpp_type=1,
            label=1,
            has_default_value=False,
            default_value=0,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            options=None,
        ),
        _descriptor.FieldDescriptor(
            name="height",
            full_name="Datum.height",
            index=1,
            number=2,
            type=5,
            cpp_type=1,
            label=1,
            has_default_value=False,
            default_value=0,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            options=None,
        ),
        _descriptor.FieldDescriptor(
            name="width",
            full_name="Datum.width",
            index=2,
            number=3,
            type=5,
            cpp_type=1,
            label=1,
            has_default_value=False,
            default_value=0,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            options=None,
        ),
        _descriptor.FieldDescriptor(
            name="data",
            full_name="Datum.data",
            index=3,
            number=4,
            type=12,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=_b(""),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            options=None,
        ),
        _descriptor.FieldDescriptor(
            name="label",
            full_name="Datum.label",
            index=4,
            number=5,
            type=5,
            cpp_type=1,
            label=1,
            has_default_value=False,
            default_value=0,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            options=None,
        ),
        _descriptor.FieldDescriptor(
            name="float_data",
            full_name="Datum.float_data",
            index=5,
            number=6,
            type=2,
            cpp_type=6,
            label=3,
            has_default_value=False,
            default_value=[],
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            options=None,
        ),
        _descriptor.FieldDescriptor(
            name="encoded",
            full_name="Datum.encoded",
            index=6,
            number=7,
            type=8,
            cpp_type=7,
            label=1,
            has_default_value=True,
            default_value=False,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            options=None,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    options=None,
    is_extendable=False,
    extension_ranges=[],
    oneofs=[],
    serialized_start=16,
    serialized_end=145,
)

DESCRIPTOR.message_types_by_name["Datum"] = _DATUM

Datum = _reflection.GeneratedProtocolMessageType(
    "Datum",
    (_message.Message,),
    dict(
        DESCRIPTOR=_DATUM,
        __module__="datum_pb2"
        # @@protoc_insertion_point(class_scope:Datum)
    ),
)
_sym_db.RegisterMessage(Datum)


# @@protoc_insertion_point(module_scope)



import logging
import math

import slidingwindow as sw

import cv2
import numpy as np
import tensorflow as tf
import time

from tf_pose import common
from tf_pose.common import CocoPart
from tf_pose.tensblur.smoother import Smoother
import tensorflow.contrib.tensorrt as trt

try:
    from tf_pose.pafprocess import pafprocess
except ModuleNotFoundError as e:
    print(e)
    print(
        "you need to build c++ library for pafprocess. See : https://github.com/ildoonet/tf-pose-estimation/tree/master/tf_pose/pafprocess"
    )
    exit(-1)

logger = logging.getLogger("TfPoseEstimator")
logger.handlers.clear()
logger.setLevel(logging.INFO)
ch = logging.StreamHandler()
formatter = logging.Formatter("[%(asctime)s] [%(name)s] [%(levelname)s] %(message)s")
ch.setFormatter(formatter)
logger.addHandler(ch)
logger.setLevel(logging.INFO)


def _round(v):
    return int(round(v))


def _include_part(part_list, part_idx):
    for part in part_list:
        if part_idx == part.part_idx:
            return True, part
    return False, None


class Human:
    """
    body_parts: list of BodyPart
    """

    __slots__ = ("body_parts", "pairs", "uidx_list", "score")

    def __init__(self, pairs):
        self.pairs = []
        self.uidx_list = set()
        self.body_parts = {}
        for pair in pairs:
            self.add_pair(pair)
        self.score = 0.0

    @staticmethod
    def _get_uidx(part_idx, idx):
        return "%d-%d" % (part_idx, idx)

    def add_pair(self, pair):
        self.pairs.append(pair)
        self.body_parts[pair.part_idx1] = BodyPart(
            Human._get_uidx(pair.part_idx1, pair.idx1),
            pair.part_idx1,
            pair.coord1[0],
            pair.coord1[1],
            pair.score,
        )
        self.body_parts[pair.part_idx2] = BodyPart(
            Human._get_uidx(pair.part_idx2, pair.idx2),
            pair.part_idx2,
            pair.coord2[0],
            pair.coord2[1],
            pair.score,
        )
        self.uidx_list.add(Human._get_uidx(pair.part_idx1, pair.idx1))
        self.uidx_list.add(Human._get_uidx(pair.part_idx2, pair.idx2))

    def is_connected(self, other):
        return len(self.uidx_list & other.uidx_list) > 0

    def merge(self, other):
        for pair in other.pairs:
            self.add_pair(pair)

    def part_count(self):
        return len(self.body_parts.keys())

    def get_max_score(self):
        return max([x.score for _, x in self.body_parts.items()])

    def get_face_box(self, img_w, img_h, mode=0):
        """
        Get Face box compared to img size (w, h)
        :param img_w:
        :param img_h:
        :param mode:
        :return:
        """
        # SEE : https://github.com/ildoonet/tf-pose-estimation/blob/master/tf_pose/common.py#L13
        _NOSE = CocoPart.Nose.value
        _NECK = CocoPart.Neck.value
        _REye = CocoPart.REye.value
        _LEye = CocoPart.LEye.value
        _REar = CocoPart.REar.value
        _LEar = CocoPart.LEar.value

        _THRESHOLD_PART_CONFIDENCE = 0.2
        parts = [
            part
            for idx, part in self.body_parts.items()
            if part.score > _THRESHOLD_PART_CONFIDENCE
        ]

        is_nose, part_nose = _include_part(parts, _NOSE)
        if not is_nose:
            return None

        size = 0
        is_neck, part_neck = _include_part(parts, _NECK)
        if is_neck:
            size = max(size, img_h * (part_neck.y - part_nose.y) * 0.8)

        is_reye, part_reye = _include_part(parts, _REye)
        is_leye, part_leye = _include_part(parts, _LEye)
        if is_reye and is_leye:
            size = max(size, img_w * (part_reye.x - part_leye.x) * 2.0)
            size = max(
                size,
                img_w
                * math.sqrt(
                    (part_reye.x - part_leye.x) ** 2 + (part_reye.y - part_leye.y) ** 2
                )
                * 2.0,
            )

        if mode == 1:
            if not is_reye and not is_leye:
                return None

        is_rear, part_rear = _include_part(parts, _REar)
        is_lear, part_lear = _include_part(parts, _LEar)
        if is_rear and is_lear:
            size = max(size, img_w * (part_rear.x - part_lear.x) * 1.6)

        if size <= 0:
            return None

        if not is_reye and is_leye:
            x = part_nose.x * img_w - (size // 3 * 2)
        elif is_reye and not is_leye:
            x = part_nose.x * img_w - (size // 3)
        else:  # is_reye and is_leye:
            x = part_nose.x * img_w - size // 2

        x2 = x + size
        if mode == 0:
            y = part_nose.y * img_h - size // 3
        else:
            y = part_nose.y * img_h - _round(size / 2 * 1.2)
        y2 = y + size

        # fit into the image frame
        x = max(0, x)
        y = max(0, y)
        x2 = min(img_w - x, x2 - x) + x
        y2 = min(img_h - y, y2 - y) + y

        if _round(x2 - x) == 0.0 or _round(y2 - y) == 0.0:
            return None
        if mode == 0:
            return {
                "x": _round((x + x2) / 2),
                "y": _round((y + y2) / 2),
                "w": _round(x2 - x),
                "h": _round(y2 - y),
            }
        else:
            return {
                "x": _round(x),
                "y": _round(y),
                "w": _round(x2 - x),
                "h": _round(y2 - y),
            }

    def get_upper_body_box(self, img_w, img_h):
        """
        Get Upper body box compared to img size (w, h)
        :param img_w:
        :param img_h:
        :return:
        """

        if not (img_w > 0 and img_h > 0):
            raise Exception("img size should be positive")

        _NOSE = CocoPart.Nose.value
        _NECK = CocoPart.Neck.value
        _RSHOULDER = CocoPart.RShoulder.value
        _LSHOULDER = CocoPart.LShoulder.value
        _THRESHOLD_PART_CONFIDENCE = 0.3
        parts = [
            part
            for idx, part in self.body_parts.items()
            if part.score > _THRESHOLD_PART_CONFIDENCE
        ]
        part_coords = [
            (img_w * part.x, img_h * part.y)
            for part in parts
            if part.part_idx in [0, 1, 2, 5, 8, 11, 14, 15, 16, 17]
        ]

        if len(part_coords) < 5:
            return None

        # Initial Bounding Box
        x = min([part[0] for part in part_coords])
        y = min([part[1] for part in part_coords])
        x2 = max([part[0] for part in part_coords])
        y2 = max([part[1] for part in part_coords])

        # # ------ Adjust heuristically +
        # if face points are detcted, adjust y value

        is_nose, part_nose = _include_part(parts, _NOSE)
        is_neck, part_neck = _include_part(parts, _NECK)
        torso_height = 0
        if is_nose and is_neck:
            y -= (part_neck.y * img_h - y) * 0.8
            torso_height = max(0, (part_neck.y - part_nose.y) * img_h * 2.5)
        #
        # # by using shoulder position, adjust width
        is_rshoulder, part_rshoulder = _include_part(parts, _RSHOULDER)
        is_lshoulder, part_lshoulder = _include_part(parts, _LSHOULDER)
        if is_rshoulder and is_lshoulder:
            half_w = x2 - x
            dx = half_w * 0.15
            x -= dx
            x2 += dx
        elif is_neck:
            if is_lshoulder and not is_rshoulder:
                half_w = abs(part_lshoulder.x - part_neck.x) * img_w * 1.15
                x = min(part_neck.x * img_w - half_w, x)
                x2 = max(part_neck.x * img_w + half_w, x2)
            elif not is_lshoulder and is_rshoulder:
                half_w = abs(part_rshoulder.x - part_neck.x) * img_w * 1.15
                x = min(part_neck.x * img_w - half_w, x)
                x2 = max(part_neck.x * img_w + half_w, x2)

        # ------ Adjust heuristically -

        # fit into the image frame
        x = max(0, x)
        y = max(0, y)
        x2 = min(img_w - x, x2 - x) + x
        y2 = min(img_h - y, y2 - y) + y

        if _round(x2 - x) == 0.0 or _round(y2 - y) == 0.0:
            return None
        return {
            "x": _round((x + x2) / 2),
            "y": _round((y + y2) / 2),
            "w": _round(x2 - x),
            "h": _round(y2 - y),
        }

    def __str__(self):
        return " ".join([str(x) for x in self.body_parts.values()])

    def __repr__(self):
        return self.__str__()


class BodyPart:
    """
    part_idx : part index(eg. 0 for nose)
    x, y: coordinate of body part
    score : confidence score
    """

    __slots__ = ("uidx", "part_idx", "x", "y", "score")

    def __init__(self, uidx, part_idx, x, y, score):
        self.uidx = uidx
        self.part_idx = part_idx
        self.x, self.y = x, y
        self.score = score

    def get_part_name(self):
        return CocoPart(self.part_idx)

    def __str__(self):
        return "BodyPart:%d-(%.2f, %.2f) score=%.2f" % (
            self.part_idx,
            self.x,
            self.y,
            self.score,
        )

    def __repr__(self):
        return self.__str__()


class PoseEstimator:
    def __init__(self):
        pass

    @staticmethod
    def estimate_paf(peaks, heat_mat, paf_mat):
        pafprocess.process_paf(peaks, heat_mat, paf_mat)

        humans = []
        for human_id in range(pafprocess.get_num_humans()):
            human = Human([])
            is_added = False

            for part_idx in range(18):
                c_idx = int(pafprocess.get_part_cid(human_id, part_idx))
                if c_idx < 0:
                    continue

                is_added = True
                human.body_parts[part_idx] = BodyPart(
                    "%d-%d" % (human_id, part_idx),
                    part_idx,
                    float(pafprocess.get_part_x(c_idx)) / heat_mat.shape[1],
                    float(pafprocess.get_part_y(c_idx)) / heat_mat.shape[0],
                    pafprocess.get_part_score(c_idx),
                )

            if is_added:
                score = pafprocess.get_score(human_id)
                human.score = score
                humans.append(human)

        return humans


class TfPoseEstimator:
    # TODO : multi-scale

    def __init__(
        self, graph_path, target_size=(320, 240), tf_config=None, trt_bool=False
    ):
        self.target_size = target_size

        # load graph
        logger.info(
            "loading graph from %s(default size=%dx%d)"
            % (graph_path, target_size[0], target_size[1])
        )
        with tf.gfile.GFile(graph_path, "rb") as f:
            graph_def = tf.GraphDef()
            graph_def.ParseFromString(f.read())

        if trt_bool is True:
            output_nodes = ["Openpose/concat_stage7"]
            graph_def = trt.create_inference_graph(
                graph_def,
                output_nodes,
                max_batch_size=1,
                max_workspace_size_bytes=1 << 20,
                precision_mode="FP16",
                # precision_mode="INT8",
                minimum_segment_size=3,
                is_dynamic_op=True,
                maximum_cached_engines=int(1e3),
                use_calibration=True,
            )

        self.graph = tf.get_default_graph()
        tf.import_graph_def(graph_def, name="TfPoseEstimator")
        self.persistent_sess = tf.Session(graph=self.graph, config=tf_config)

        for ts in [n.name for n in tf.get_default_graph().as_graph_def().node]:
            print(ts)

        self.tensor_image = self.graph.get_tensor_by_name("TfPoseEstimator/image:0")
        self.tensor_output = self.graph.get_tensor_by_name(
            "TfPoseEstimator/Openpose/concat_stage7:0"
        )
        self.tensor_heatMat = self.tensor_output[:, :, :, :19]
        self.tensor_pafMat = self.tensor_output[:, :, :, 19:]
        self.upsample_size = tf.placeholder(
            dtype=tf.int32, shape=(2,), name="upsample_size"
        )
        self.tensor_heatMat_up = tf.image.resize_area(
            self.tensor_output[:, :, :, :19],
            self.upsample_size,
            align_corners=False,
            name="upsample_heatmat",
        )
        self.tensor_pafMat_up = tf.image.resize_area(
            self.tensor_output[:, :, :, 19:],
            self.upsample_size,
            align_corners=False,
            name="upsample_pafmat",
        )
        if trt_bool is True:
            smoother = Smoother({"data": self.tensor_heatMat_up}, 25, 3.0, 19)
        else:
            smoother = Smoother({"data": self.tensor_heatMat_up}, 25, 3.0)
        gaussian_heatMat = smoother.get_output()

        max_pooled_in_tensor = tf.nn.pool(
            gaussian_heatMat, window_shape=(3, 3), pooling_type="MAX", padding="SAME"
        )
        self.tensor_peaks = tf.where(
            tf.equal(gaussian_heatMat, max_pooled_in_tensor),
            gaussian_heatMat,
            tf.zeros_like(gaussian_heatMat),
        )

        self.heatMat = self.pafMat = None

        # warm-up
        self.persistent_sess.run(
            tf.variables_initializer(
                [
                    v
                    for v in tf.global_variables()
                    if v.name.split(":")[0]
                    in [
                        x.decode("utf-8")
                        for x in self.persistent_sess.run(
                            tf.report_uninitialized_variables()
                        )
                    ]
                ]
            )
        )
        self.persistent_sess.run(
            [self.tensor_peaks, self.tensor_heatMat_up, self.tensor_pafMat_up],
            feed_dict={
                self.tensor_image: [
                    np.ndarray(
                        shape=(target_size[1], target_size[0], 3), dtype=np.float32
                    )
                ],
                self.upsample_size: [target_size[1], target_size[0]],
            },
        )
        self.persistent_sess.run(
            [self.tensor_peaks, self.tensor_heatMat_up, self.tensor_pafMat_up],
            feed_dict={
                self.tensor_image: [
                    np.ndarray(
                        shape=(target_size[1], target_size[0], 3), dtype=np.float32
                    )
                ],
                self.upsample_size: [target_size[1] // 2, target_size[0] // 2],
            },
        )
        self.persistent_sess.run(
            [self.tensor_peaks, self.tensor_heatMat_up, self.tensor_pafMat_up],
            feed_dict={
                self.tensor_image: [
                    np.ndarray(
                        shape=(target_size[1], target_size[0], 3), dtype=np.float32
                    )
                ],
                self.upsample_size: [target_size[1] // 4, target_size[0] // 4],
            },
        )

        # logs
        if self.tensor_image.dtype == tf.quint8:
            logger.info("quantization mode enabled.")

    def __del__(self):
        # self.persistent_sess.close()
        pass

    def get_flops(self):
        flops = tf.profiler.profile(
            self.graph, options=tf.profiler.ProfileOptionBuilder.float_operation()
        )
        return flops.total_float_ops

    @staticmethod
    def _quantize_img(npimg):
        npimg_q = npimg + 1.0
        npimg_q /= 2.0 / 2 ** 8
        # npimg_q += 0.5
        npimg_q = npimg_q.astype(np.uint8)
        return npimg_q

    @staticmethod
    def draw_humans(npimg, humans, imgcopy=False):
        if imgcopy:
            npimg = np.copy(npimg)
        image_h, image_w = npimg.shape[:2]
        centers = {}
        for human in humans:
            # draw point
            for i in range(common.CocoPart.Background.value):
                if i not in human.body_parts.keys():
                    continue

                body_part = human.body_parts[i]
                center = (
                    int(body_part.x * image_w + 0.5),
                    int(body_part.y * image_h + 0.5),
                )
                centers[i] = center
                cv2.circle(
                    npimg,
                    center,
                    3,
                    common.CocoColors[i],
                    thickness=3,
                    lineType=8,
                    shift=0,
                )

            # draw line
            for pair_order, pair in enumerate(common.CocoPairsRender):
                if (
                    pair[0] not in human.body_parts.keys()
                    or pair[1] not in human.body_parts.keys()
                ):
                    continue

                # npimg = cv2.line(npimg, centers[pair[0]], centers[pair[1]], common.CocoColors[pair_order], 3)
                cv2.line(
                    npimg,
                    centers[pair[0]],
                    centers[pair[1]],
                    common.CocoColors[pair_order],
                    3,
                )

        return npimg

    def _get_scaled_img(self, npimg, scale):
        get_base_scale = (
            lambda s, w, h: max(
                self.target_size[0] / float(h), self.target_size[1] / float(w)
            )
            * s
        )
        img_h, img_w = npimg.shape[:2]

        if scale is None:
            if npimg.shape[:2] != (self.target_size[1], self.target_size[0]):
                # resize
                npimg = cv2.resize(
                    npimg, self.target_size, interpolation=cv2.INTER_CUBIC
                )
            return [npimg], [(0.0, 0.0, 1.0, 1.0)]
        elif isinstance(scale, float):
            # scaling with center crop
            base_scale = get_base_scale(scale, img_w, img_h)
            npimg = cv2.resize(
                npimg,
                dsize=None,
                fx=base_scale,
                fy=base_scale,
                interpolation=cv2.INTER_CUBIC,
            )

            o_size_h, o_size_w = npimg.shape[:2]
            if (
                npimg.shape[0] < self.target_size[1]
                or npimg.shape[1] < self.target_size[0]
            ):
                newimg = np.zeros(
                    (
                        max(self.target_size[1], npimg.shape[0]),
                        max(self.target_size[0], npimg.shape[1]),
                        3,
                    ),
                    dtype=np.uint8,
                )
                newimg[: npimg.shape[0], : npimg.shape[1], :] = npimg
                npimg = newimg

            windows = sw.generate(
                npimg,
                sw.DimOrder.HeightWidthChannel,
                self.target_size[0],
                self.target_size[1],
                0.2,
            )

            rois = []
            ratios = []
            for window in windows:
                indices = window.indices()
                roi = npimg[indices]
                rois.append(roi)
                ratio_x, ratio_y = (
                    float(indices[1].start) / o_size_w,
                    float(indices[0].start) / o_size_h,
                )
                ratio_w, ratio_h = (
                    float(indices[1].stop - indices[1].start) / o_size_w,
                    float(indices[0].stop - indices[0].start) / o_size_h,
                )
                ratios.append((ratio_x, ratio_y, ratio_w, ratio_h))

            return rois, ratios
        elif isinstance(scale, tuple) and len(scale) == 2:
            # scaling with sliding window : (scale, step)
            base_scale = get_base_scale(scale[0], img_w, img_h)
            npimg = cv2.resize(
                npimg,
                dsize=None,
                fx=base_scale,
                fy=base_scale,
                interpolation=cv2.INTER_CUBIC,
            )
            o_size_h, o_size_w = npimg.shape[:2]
            if (
                npimg.shape[0] < self.target_size[1]
                or npimg.shape[1] < self.target_size[0]
            ):
                newimg = np.zeros(
                    (
                        max(self.target_size[1], npimg.shape[0]),
                        max(self.target_size[0], npimg.shape[1]),
                        3,
                    ),
                    dtype=np.uint8,
                )
                newimg[: npimg.shape[0], : npimg.shape[1], :] = npimg
                npimg = newimg

            window_step = scale[1]

            windows = sw.generate(
                npimg,
                sw.DimOrder.HeightWidthChannel,
                self.target_size[0],
                self.target_size[1],
                window_step,
            )

            rois = []
            ratios = []
            for window in windows:
                indices = window.indices()
                roi = npimg[indices]
                rois.append(roi)
                ratio_x, ratio_y = (
                    float(indices[1].start) / o_size_w,
                    float(indices[0].start) / o_size_h,
                )
                ratio_w, ratio_h = (
                    float(indices[1].stop - indices[1].start) / o_size_w,
                    float(indices[0].stop - indices[0].start) / o_size_h,
                )
                ratios.append((ratio_x, ratio_y, ratio_w, ratio_h))

            return rois, ratios
        elif isinstance(scale, tuple) and len(scale) == 3:
            # scaling with ROI : (want_x, want_y, scale_ratio)
            base_scale = get_base_scale(scale[2], img_w, img_h)
            npimg = cv2.resize(
                npimg,
                dsize=None,
                fx=base_scale,
                fy=base_scale,
                interpolation=cv2.INTER_CUBIC,
            )
            ratio_w = self.target_size[0] / float(npimg.shape[1])
            ratio_h = self.target_size[1] / float(npimg.shape[0])

            want_x, want_y = scale[:2]
            ratio_x = want_x - ratio_w / 2.0
            ratio_y = want_y - ratio_h / 2.0
            ratio_x = max(ratio_x, 0.0)
            ratio_y = max(ratio_y, 0.0)
            if ratio_x + ratio_w > 1.0:
                ratio_x = 1.0 - ratio_w
            if ratio_y + ratio_h > 1.0:
                ratio_y = 1.0 - ratio_h

            roi = self._crop_roi(npimg, ratio_x, ratio_y)
            return [roi], [(ratio_x, ratio_y, ratio_w, ratio_h)]

    def _crop_roi(self, npimg, ratio_x, ratio_y):
        target_w, target_h = self.target_size
        h, w = npimg.shape[:2]
        x = max(int(w * ratio_x - 0.5), 0)
        y = max(int(h * ratio_y - 0.5), 0)
        cropped = npimg[y : y + target_h, x : x + target_w]

        cropped_h, cropped_w = cropped.shape[:2]
        if cropped_w < target_w or cropped_h < target_h:
            npblank = np.zeros(
                (self.target_size[1], self.target_size[0], 3), dtype=np.uint8
            )

            copy_x, copy_y = (target_w - cropped_w) // 2, (target_h - cropped_h) // 2
            npblank[copy_y : copy_y + cropped_h, copy_x : copy_x + cropped_w] = cropped
        else:
            return cropped

    def inference(self, npimg, resize_to_default=True, upsample_size=1.0):
        if npimg is None:
            raise Exception("The image is not valid. Please check your image exists.")

        if resize_to_default:
            upsample_size = [
                int(self.target_size[1] / 8 * upsample_size),
                int(self.target_size[0] / 8 * upsample_size),
            ]
        else:
            upsample_size = [
                int(npimg.shape[0] / 8 * upsample_size),
                int(npimg.shape[1] / 8 * upsample_size),
            ]

        if self.tensor_image.dtype == tf.quint8:
            # quantize input image
            npimg = TfPoseEstimator._quantize_img(npimg)
            pass

        logger.debug(
            "inference+ original shape=%dx%d" % (npimg.shape[1], npimg.shape[0])
        )
        img = npimg
        if resize_to_default:
            img = self._get_scaled_img(npimg, None)[0][0]
        peaks, heatMat_up, pafMat_up = self.persistent_sess.run(
            [self.tensor_peaks, self.tensor_heatMat_up, self.tensor_pafMat_up],
            feed_dict={self.tensor_image: [img], self.upsample_size: upsample_size},
        )
        peaks = peaks[0]
        self.heatMat = heatMat_up[0]
        self.pafMat = pafMat_up[0]
        logger.debug(
            "inference- heatMat=%dx%d pafMat=%dx%d"
            % (
                self.heatMat.shape[1],
                self.heatMat.shape[0],
                self.pafMat.shape[1],
                self.pafMat.shape[0],
            )
        )

        t = time.time()
        humans = PoseEstimator.estimate_paf(peaks, self.heatMat, self.pafMat)
        logger.debug("estimate time=%.5f" % (time.time() - t))
        return humans


if __name__ == "__main__":
    import pickle

    f = open("./etcs/heatpaf1.pkl", "rb")
    data = pickle.load(f)
    logger.info("size={}".format(data["heatMat"].shape))
    f.close()

    t = time.time()
    humans = PoseEstimator.estimate_paf(data["peaks"], data["heatMat"], data["pafMat"])
    dt = time.time() - t
    t = time.time()
    logger.info("elapsed #humans=%d time=%.8f" % (len(humans), dt))



import math, mmap, tempfile
import numpy as np
import psutil


def _requiredSize(shape, dtype):
    """
	Determines the number of bytes required to store a NumPy array with
	the specified shape and datatype.
	"""
    return math.floor(
        np.prod(np.asarray(shape, dtype=np.uint64)) * np.dtype(dtype).itemsize
    )


class TempfileBackedArray(np.ndarray):
    """
	A NumPy ndarray that uses a memory-mapped temp file as its backing 
	"""

    def __new__(
        subtype,
        shape,
        dtype=float,
        buffer=None,
        offset=0,
        strides=None,
        order=None,
        info=None,
    ):

        # Determine the size in bytes required to hold the array
        numBytes = _requiredSize(shape, dtype)

        # Create the temporary file, resize it, and map it into memory
        tempFile = tempfile.TemporaryFile()
        tempFile.truncate(numBytes)
        buf = mmap.mmap(tempFile.fileno(), numBytes, access=mmap.ACCESS_WRITE)

        # Create the ndarray with the memory map as the underlying buffer
        obj = super(TempfileBackedArray, subtype).__new__(
            subtype, shape, dtype, buf, 0, None, order
        )

        # Attach the file reference to the ndarray object
        obj._file = tempFile
        return obj

    def __array_finalize__(self, obj):
        if obj is None:
            return
        self._file = getattr(obj, "_file", None)


def arrayFactory(shape, dtype=float):
    """
	Creates a new ndarray of the specified shape and datatype, storing
	it in memory if there is sufficient available space or else using
	a memory-mapped temporary file to provide the underlying buffer.
	"""

    # Determine the number of bytes required to store the array
    requiredBytes = _requiredSize(shape, dtype)

    # Determine if there is sufficient available memory
    vmem = psutil.virtual_memory()
    if vmem.available > requiredBytes:
        return np.ndarray(shape=shape, dtype=dtype)
    else:
        return TempfileBackedArray(shape=shape, dtype=dtype)


def zerosFactory(shape, dtype=float):
    """
	Creates a new NumPy array using `arrayFactory()` and fills it with zeros.
	"""
    arr = arrayFactory(shape=shape, dtype=dtype)
    arr.fill(0)
    return arr


def arrayCast(source, dtype):
    """
	Casts a NumPy array to the specified datatype, storing the copy
	in memory if there is sufficient available space or else using a
	memory-mapped temporary file to provide the underlying buffer.
	"""

    # Determine the number of bytes required to store the array
    requiredBytes = _requiredSize(source.shape, dtype)

    # Determine if there is sufficient available memory
    vmem = psutil.virtual_memory()
    if vmem.available > requiredBytes:
        return source.astype(dtype, subok=False)
    else:
        dest = arrayFactory(source.shape, dtype)
        np.copyto(dest, source, casting="unsafe")
        return dest


def determineMaxWindowSize(dtype, limit=None):
    """
	Determines the largest square window size that can be used, based on
	the specified datatype and amount of currently available system memory.
	
	If `limit` is specified, then this value will be returned in the event
	that it is smaller than the maximum computed size.
	"""
    vmem = psutil.virtual_memory()
    maxSize = math.floor(math.sqrt(vmem.available / np.dtype(dtype).itemsize))
    if limit is None or limit >= maxSize:
        return maxSize
    else:
        return limit



from .SlidingWindow import generate
from .Batching import batchWindows
import numpy as np


def mergeWindows(
    data,
    dimOrder,
    maxWindowSize,
    overlapPercent,
    batchSize,
    transform,
    progressCallback=None,
):
    """
	Generates sliding windows for the specified dataset and applies the specified
	transformation function to each window. Where multiple overlapping windows
	include an element of the input dataset, the overlap is resolved by computing
	the mean transform result value for that element.
	
	Irrespective of the order of the dimensions of the input dataset, the
	transformation function should return a NumPy array with dimensions
	[batch, height, width, resultChannels].
	
	If a progress callback is supplied, it will be called immediately before
	applying the transformation function to each batch of windows. The callback
	should accept the current batch index and number of batches as arguments.
	"""

    # Determine the dimensions of the input data
    sourceWidth = data.shape[dimOrder.index("w")]
    sourceHeight = data.shape[dimOrder.index("h")]

    # Generate the sliding windows and group them into batches
    windows = generate(data, dimOrder, maxWindowSize, overlapPercent)
    batches = batchWindows(windows, batchSize)

    # Apply the transform to the first batch of windows and determine the result dimensionality
    exemplarResult = transform(data, batches[0])
    resultDimensions = exemplarResult.shape[len(exemplarResult.shape) - 1]

    # Create the matrices to hold the sums and counts for the transform result values
    sums = np.zeros((sourceHeight, sourceWidth, resultDimensions), dtype=np.float)
    counts = np.zeros((sourceHeight, sourceWidth), dtype=np.uint32)

    # Iterate over the batches and apply the transformation function to each batch
    for batchNum, batch in enumerate(batches):

        # If a progress callback was supplied, call it
        if progressCallback != None:
            progressCallback(batchNum, len(batches))

        # Apply the transformation function to the current batch
        batchResult = transform(data, batch)

        # Iterate over the windows in the batch and update the sums matrix
        for windowNum, window in enumerate(batch):

            # Create views into the larger matrices that correspond to the current window
            windowIndices = window.indices(False)
            sumsView = sums[windowIndices]
            countsView = counts[windowIndices]

            # Update the result sums for each of the dataset elements in the window
            sumsView[:] += batchResult[windowNum]
            countsView[:] += 1

    # Use the sums and the counts to compute the mean values
    for dim in range(0, resultDimensions):
        sums[:, :, dim] /= counts

    # Return the mean values
    return sums
